# gophian -- tools to help with Debianizing Go software
# Copyright (C) 2024 Maytham Alsudany <maytha8thedev@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Optional

import click
import graphviz
from graphviz.dot import Dot
import requests

from gophian.__about__ import BUGS_URL
from gophian.error import ExecutionError
from gophian.hosts import KNOWN_HOSTS
from gophian.name import debianize_name, shorten_host
from gophian.packages import DebianGolangPackages
from gophian.session import Session
from gophian.vcs import package_from_import_path


@click.command()
@click.argument("importpath")
@click.argument("output", type=click.Path(exists=False, writable=True), required=False)
@click.option("--quiet/--no-quiet", default=False, help="Don't print warnings.")
@click.option(
    "--open/--no-open",
    "should_open",
    default=True,
    help="View the rendered output in the most suitable application.",
)
@click.option(
    "--render/--no-render",
    default=True,
    help="Render the graph using graphviz.",
)
@click.option(
    "--warn-packaged/--no-warn-packaged",
    default=True,
    help="Warn if the given Go package has already been packaged for Debian.",
)
@click.option(
    "--include-packaged-deps/--no-include-packaged-deps",
    default=True,
    help="Include packaged dependencies in the estimate.",
)
@click.pass_context
def estimate_graphviz(
    context: click.Context,
    importpath: str,
    output: Optional[str],
    quiet: bool,
    should_open: bool,
    render: bool,
    warn_packaged: bool,
    include_packaged_deps: bool,
) -> None:
    """
    Visualize the dependency tree of a given Go package.
    """

    requests_session: requests.Session = context.obj

    if not quiet:
        click.secho("gophian is experimental software!", bold=True, fg="yellow")
        click.secho("Please report any problems to:", fg="yellow")
        click.secho(BUGS_URL, fg="yellow")

    if render and output == "-":
        click.secho(
            "Invalid output argument. Cannot pipe rendered PDF to stdout, please specify a file name.",
            fg="red",
        )
        context.exit(1)

    try:
        package, _ = package_from_import_path(requests_session, importpath)
    except Exception as error:
        click.echo(error)
        click.secho("Did you specify a Go package import path?", fg="yellow")
        context.exit(1)

    if package != importpath:
        click.echo(f"Continuing with repo root {package} instead of {importpath}")

    debian_packages = DebianGolangPackages(requests_session)

    if debian_packages._check_for_package(package, quiet) and warn_packaged:
        if not quiet:
            click.secho("To ignore, pass the --ignore-packaged flag.", fg="cyan")
        return

    with Session(requests_session) as session:
        seen: List[str] = []
        errors: List[ExecutionError] = []

        dot = graphviz.Digraph()
        dot.attr("graph", ratio="fill", size="180,30!", margin="0")
        dot.node(
            "comment",
            "Node colours\\nWhite: not packaged\\nGreen: packaged",
            shape="none",
        )
        dot.node(package, package)

        dependency_tree(
            session,
            seen,
            errors,
            package,
            include_packaged_deps,
            debian_packages,
            dot,
        )

        if not output:
            host = package.split("/")[0]
            short_host = (
                KNOWN_HOSTS[host] if host in KNOWN_HOSTS else shorten_host(host)
            )
            name = debianize_name(package, short_host)
            output = name + ".pdf"
        source_output = output + ".gv" if render else output

        if render:
            dot.render(source_output, view=should_open, format="pdf", outfile=output)
        else:
            with click.open_file(source_output, "x") as file:
                file.write(str(dot))

        for error in errors:
            click.echo(error)


def dependency_tree(
    session: Session,
    seen: List[str],
    errors: List[ExecutionError],
    package: str,
    include_packaged_deps: bool,
    debian_packages: DebianGolangPackages,
    dot: Dot,
    level: int = 1,
):
    try:
        dep_packages = session.package_deps(package)
    except ExecutionError as error:
        errors.append(error)
        click.secho("Error getting dependencies for " + package, fg="red")
        dot.node(package, package + "\\n(Error getting dependencies)")
        return
    for dep_package, _ in dep_packages:
        if debian_package := debian_packages.library_is_packaged(dep_package):
            seen.append(dep_package)
            if include_packaged_deps:
                if dep_package not in seen:
                    dot.node(
                        dep_package,
                        f"{dep_package}\\n({debian_package})",
                        fillcolor="green",
                        style="filled",
                    )
                dot.edge(package, dep_package)
        elif dep_package in seen:
            dot.edge(package, dep_package)
        else:
            dot.node(dep_package, fillcolor="white", style="filled")
            dot.edge(package, dep_package)
            seen.append(dep_package)
            dependency_tree(
                session,
                seen,
                errors,
                dep_package,
                include_packaged_deps,
                debian_packages,
                dot,
                level + 1,
            )
